import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public
from nomad.datamodel.metainfo import common

m_package = Package(
    name='lammps_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='lammps.nomadmetainfo.json'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_lammps_atom_positions_image_index = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        PBC image flag index.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_atom_positions_image_index'))

    x_lammps_atom_positions_scaled = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        Position of the atoms in a scaled format [0, 1].
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_atom_positions_scaled'))

    x_lammps_atom_positions_wrapped = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Position of the atoms wrapped back to the periodic box.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_atom_positions_wrapped'))

    x_lammps_trj_timestep_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_trj_timestep_store'))

    x_lammps_trj_number_of_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_trj_number_of_atoms_store'))

    x_lammps_trj_box_bound_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_trj_box_bound_store'))

    x_lammps_trj_box_bounds_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_trj_box_bounds_store'))

    x_lammps_trj_variables_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_trj_variables_store'))

    x_lammps_trj_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_trj_atoms_store'))


class section_sampling_method(public.section_sampling_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_sampling_method'))

    x_lammps_barostat_target_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        MD barostat target pressure.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_barostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_lammps_barostat_target_pressure'))

    x_lammps_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD barostat relaxation time.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_barostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_lammps_barostat_tau'))

    x_lammps_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD barostat type, valid values are defined in the barostat_type wiki page.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_barostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_lammps_barostat_type'))

    x_lammps_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD integration time step.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_integrator],
        a_legacy=LegacyDefinition(name='x_lammps_integrator_dt'))

    x_lammps_integrator_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD integrator type, valid values are defined in the integrator_type wiki page.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_integrator],
        a_legacy=LegacyDefinition(name='x_lammps_integrator_type'))

    x_lammps_langevin_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Langevin thermostat damping factor.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_thermostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_lammps_langevin_gamma'))

    x_lammps_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of requested MD integration time steps.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_integrator],
        a_legacy=LegacyDefinition(name='x_lammps_number_of_steps_requested'))

    x_lammps_thermostat_level = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat level (see wiki: single, multiple, regional).
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_thermostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_lammps_thermostat_level'))

    x_lammps_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        MD thermostat target temperature.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_thermostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_lammps_thermostat_target_temperature'))

    x_lammps_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD thermostat relaxation time.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_thermostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_lammps_thermostat_tau'))

    x_lammps_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat type, valid values are defined in the thermostat_type wiki page.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_thermostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_lammps_thermostat_type'))


class section_interaction(common.section_interaction):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_interaction'))

    x_lammps_interaction_atom_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each interaction atoms.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_interaction_atom_to_atom_type_ref'))

    x_lammps_number_of_defined_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions (L-J pairs).
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_number_of_defined_pair_interactions'))

    x_lammps_pair_interaction_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['x_lammps_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_pair_interaction_atom_type_ref'))

    x_lammps_pair_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['x_lammps_number_of_defined_pair_interactions', 2],
        description='''
        Pair interactions parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_pair_interaction_parameters'))


class section_molecule_interaction(common.section_molecule_interaction):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_molecule_interaction'))

    x_lammps_molecule_interaction_atom_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each molecule interaction atoms.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_molecule_interaction_atom_to_atom_type_ref'))

    x_lammps_number_of_defined_molecule_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions within a molecule (L-J pairs).
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_number_of_defined_molecule_pair_interactions'))

    x_lammps_pair_molecule_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_defined_molecule_pair_interactions', 2],
        description='''
        Molecule pair interactions parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_pair_molecule_interaction_parameters'))

    x_lammps_pair_molecule_interaction_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['x_lammps_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions within a molecule.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_pair_molecule_interaction_to_atom_type_ref'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_lammps_program_version_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program version date.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_program_version_date'))

    x_lammps_dummy_text = Quantity(
        type=str,
        shape=[],
        description='''
        dummy text
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_dummy_text'))


class section_topology(common.section_topology):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_topology'))

    x_lammps_xlo_xhi = Quantity(
        type=str,
        shape=[],
        description='''
        test
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_xlo_xhi'))

    x_lammps_data_file_store = Quantity(
        type=str,
        shape=[],
        description='''
        Filename of data file
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_file_store'))

    x_lammps_dummy = Quantity(
        type=str,
        shape=[],
        description='''
        dummy
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_dummy'))

    x_lammps_input_units_store = Quantity(
        type=str,
        shape=[],
        description='''
        It determines the units of all quantities specified in the input script and data
        file, as well as quantities output to the screen, log file, and dump files.
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_input_units_store'))

    x_lammps_data_bd_types_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_bd_types_store'))

    x_lammps_data_bd_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_bd_count_store'))

    x_lammps_data_ag_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_ag_count_store'))

    x_lammps_data_at_types_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_at_types_store'))

    x_lammps_data_dh_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_dh_count_store'))

    x_lammps_data_angles_store = Quantity(
        type=str,
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_angles_store'))

    x_lammps_data_angle_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_angle_list_store'))

    x_lammps_data_bond_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_bond_list_store'))

    x_lammps_data_dihedral_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_dihedral_list_store'))

    x_lammps_data_dihedral_coeff_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_dihedral_coeff_list_store'))

    x_lammps_masses_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_masses_store'))

    x_lammps_data_topo_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_lammps_data_topo_list_store'))


m_package.__init_metainfo__()



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import setup_paths

import logging
import os


from LammpsCommon import get_metaInfo
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.simple_parser import mainFunction, AncillaryParser
import LammpsDataParser

nformat = {
    'float': r"[-+]?\d*\.\d+",
    'int': r"[-+]?\d+",
    'file': r"[0-9a-zA-Z_.]+"
}



# import numpy as np
# import setup_paths
# import logging, os, re, sys
# import nomadcore.ActivateLogging

############################################################
# This is the parser for the input file of LAMMPS.
############################################################

logger = logging.getLogger(name="nomad.LammpsInputParser")

# LAMMPS input script
# (1) LAMMPS does not read your entire input script and then perform a simulation with all the settings. Rather, the input script is read one line at a time and each command takes effect when it is read.
# default values ????

# Parsing rules: (source: http://lammps.sandia.gov/doc/Section_commands.html#parsing-rules)
# (1) If the last printable character on the line is a "&" character, the command is assumed to continue on the next line.
# (2) All characters from the first "#" character onward are treated as comment and discarded.
# (3) The line is searched repeatedly for $ characters, which indicate variables that are replaced with a text string.
# (4) The line is broken into "words" separated by whitespace (tabs, spaces). Note that words can thus contain letters, digits, underscores, or punctuation characters.
# (5) The first word is the command name. All successive words in the line are arguments.
# (6) If you want text with spaces to be treated as a single argument, it can be enclosed in either single or double or triple quotes.
#
# A LAMMPS input script typically has 4 parts:
#
# (1) Initialization
# (2) Atom definition
# (3) Settings
# (4) Run a simulation
#


def assignMolecules(bond_list, topo_list, at_count):  # FINDING INDIVIDUAL MOLECULES FROM BONDING PATTERN

    store = []
    for i in range(len(bond_list)):
        at1   = int(bond_list[i][2])  # atom 1 index
        at2   = int(bond_list[i][3])  # atom 2 index
        store.append([at1, at2])


    storeLine = []
    for i in range(len(bond_list)):
        at1   = int(bond_list[i][2])  # atom 1 index
        at2   = int(bond_list[i][3])  # atom 2 index
        storeLine.append(at1)
        storeLine.append(at2)
        #print len(store)

    ### TOPOLOGY'S ATOM TYPE PATTERN ###
    atomTypeInTopology = []
    for i in range(len(topo_list)):
        temp = int(topo_list[i][2])
        atomTypeInTopology.append(temp) # Atom type pattern throughout topology

    topologyPattern = " ".join([ str(x) for x in atomTypeInTopology])
    #print topologyPattern
    ############

    #####################################################################
    ###### FIND ALL MOLECULE TYPES AND ALL INDIVIDUAL MOLECULES #########
    #####################################################################

    goon             = True # looping over all covalent bonds
    nextMolecule     = 0
    moleculeId       = 0    # molecular index
    moleculePattern  = str  # atom type pattern as string (useful for matching/counting)
    moleculeInfo     = []   # list containing molecule info


    while goon == True:
        atomIndexInMolecule = [] # list storing the index of atoms within that molecule
        atomTypeInMolecule  = [] # list storing the type of atoms within that molecule

        atomPositionInMolecule = [] # list storing the molecular atom index, example: OHH -> 123
        atomCount = 0

        for i in range(nextMolecule, len(store)):
            atomIndexInMolecule.append(store[i][0])
            atomIndexInMolecule.append(store[i][1])

            try:
                if store[i+1][0] not in atomIndexInMolecule and store[i+1][1] not in atomIndexInMolecule:
                    moleculeId += 1
                    #print moleculeId
                    break
            except IndexError:
                moleculeId += 1
                #print moleculeId
                break


        atomIndexToTopology = atomIndexInMolecule
        nextMolecule += len(atomIndexToTopology) // 2 ######## NOTA BENE ## I will start from this  to find the next molecule in the list "store"
        #print nextMolecule

        atomIndexInMolecule = sorted(list(set(atomIndexInMolecule)))  # clear duplicates and return the list of atom indexes in the molecule

        for i in atomIndexInMolecule:
            temp = int(topo_list[i-1][2])
            atomTypeInMolecule.append(temp)

            atomCount += 1
            atomPositionInMolecule.append(atomCount)

        moleculePattern = " ".join([ str(x) for x in atomTypeInMolecule])

        newMolecule = [ moleculeId, atomIndexInMolecule, atomTypeInMolecule, atomPositionInMolecule ] # storing molecule information
        moleculeInfo.append(newMolecule)

        #print atomIndexInMolecule
        #print atomTypeInMolecule

        if nextMolecule == len(store):
            goon = False

        #### Storing molecule info for each molecule type (the size of moleculeTypeInfo is the number of molecule types)
    moleculeTypeInfo = []
    ghost = []
    for line in moleculeInfo:
        seen = line[2]
        temp = [line[0], line[2], line[1]]

        if seen not in ghost:
            ghost.append(seen)
            moleculeTypeInfo.append(temp)
        #################

    for i in range(len(moleculeTypeInfo)):
        for j in range(len(moleculeInfo)):

            if moleculeTypeInfo[i][1] == moleculeInfo[j][2]:
                moleculeInfo[j].insert(1, i+1)   ### moleculeInfo contains: [ moleculeId, moleculeType, atomIndexInMolecule,
                ###                          atomTypeInMolecule, atomPositionInMolecule ]

    atomPositionInMoleculeList = []
    for i in range(len(moleculeInfo)):
        atomPositionInMoleculeList = atomPositionInMoleculeList + moleculeInfo[i][4] # complete list storing the molecular atom index, example: OHHOHHOHH -> 123123123


    moleculeInfoResolved = []
    for i in range(0, at_count):
        for j in range(len(moleculeInfo)):

            if i+1 in moleculeInfo[j][2]:
                moleculeInfoResolved.append([ i+1, moleculeInfo[j][0], moleculeInfo[j][1], atomPositionInMoleculeList[i] ])


    # print moleculeTypeInfo

    return (moleculeTypeInfo, moleculeInfo, moleculeInfoResolved)







class LammpsInputParserContext(object):
    """Context for parsing LAMMPS input file.

    Attributes:
        # dos_energies: Stores parsed energies.
        # dos_values: Stores parsed DOS values.

    This class keeps tracks of several lammps settings to adjust the parsing to them.
    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of writing and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """

    def __init__(self):
        pass



    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts.

        Get compiled parser, filename and metadata.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        self.fName = fInName

        # save metadata
        self.metaInfoEnv = self.parser.parserBuilder.metaInfoEnv

        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()


    def initialize_values(self):
        """Initializes the values of certain variables.

        This allows a consistent setting and resetting of the variables,
        when the parsing starts and when a section_run closes.
        """
        self.units='lj'



    def compile_data_parser(self):
        """Instantiate superContext and construct parser for external data file.
        """
        self.dataSuperContext = LammpsDataParser.LammpsDataParserContext(False)
        self.dataParser = AncillaryParser(
            fileDescription = LammpsDataParser.build_LammpsDataFileSimpleMatcher(),
            parser = self.parser,
            cachingLevelForMetaName = LammpsDataParser.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.Ignore),
            superContext = self.dataSuperContext
        )

    def onClose_section_run(self, backend, gIndex, section):
        pass

    def onClose_section_topology(self, backend, gIndex, section):
        """Trigger called when section_topology is closed.

        Input data is parsed from external file but the result is only stored and written later in lammps_topology
        in section_run.
        """

        tmp = section['x_lammps_dummy_text']


        if section['x_lammps_input_units_store'] is not None:
            self.units = section['x_lammps_input_units_store']


        # construct file name
        data_file = section['x_lammps_data_file_store']

        if data_file is not None:
            self.compile_data_parser()

            dirName = os.path.dirname(os.path.abspath(self.fName))
            fName = os.path.normpath(os.path.join(dirName, data_file[0]))

            try:
                with open(fName) as fIn:
                    # construct parser for DOS file if not present
                    if self.dataSuperContext is None or self.dataParser is None:
                        self.compile_data_parser()
                    # parse DOS file
                    self.dataParser.parseFile(fIn)
                    pass
                    # set flag that DOS was parsed successfully and store values
                    # if self.dosSuperContext.dos_energies is not None and self.dosSuperContext.dos_values is not None:
                    #     self.dosFound = True
                    #     self.dos_energies = self.dosSuperContext.dos_energies
                    #     self.dos_values = self.dosSuperContext.dos_values
                    # else:
                    #     logger.error("DOS parsing unsuccessful. Parsing of file %s in directory '%s' did not yield energies or values for DOS." % (data_file, dirName))
            except IOError:
                logger.error("DOS parsing unsuccessful. Could not find %s file in directory '%s'." % (data_file, dirName))

    def onClose_section_dos(self, backend, gIndex, section):
        """Trigger called when section_dos is closed.

        Store the parsed values and write them if writeMetaData is True.
        """
        # extract energies
        dos_energies = section['fhi_aims_dos_energy']
        if dos_energies is not None:
            pass
            # self.dos_energies = np.asarray(dos_energies)
        # extract dos values
        dos_values = []
        if section['fhi_aims_dos_value_string'] is not None:
            for string in section['fhi_aims_dos_value_string']:
                strings = string.split()
                dos_values.append(map(float, strings))
        if dos_values:
            # need to transpose array since its shape is [number_of_spin_channels,n_dos_values] in the metadata
            pass
            # self.dos_values = np.transpose(np.asarray(dos_values))
        # write metadata only if values were found for both quantities
        if self.writeMetaData:
            if dos_energies is not None and dos_values:
                pass
                # backend.addArrayValues('dos_energies', self.dos_energies)
                # backend.addArrayValues('dos_values', self.dos_values)


def build_LammpsInputFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the DOS file of FHI-aims.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses DOS file of FHI-aims.
    """

    # A LAMMPS input script typically has 4 parts:
    #
    # (1) Initialization
    # (2) Atom definition
    # (3) Settings
    # (4) Run a simulation


    # Initialization
    # Set parameters that need to be defined before atoms are created or read-in from a file.

    # initialization = [
    #     SM(r"(?P<x_lammps_dummy_text>\s*dimension)"),
    #     SM(r"(?P<x_lammps_dummy_text>\s*atom_modify)"),
    #     SM(r"(?P<x_lammps_dummy_text>\s*atom_style)"),
    #     SM(r"(?P<x_lammps_dummy_text>\s*boundary)"),
    #     SM(r"(?P<x_lammps_dummy_text>\s*newton)"),
    #     SM(r"(?P<x_lammps_dummy_text>\s*processors)"),
    #     SM(r"(?P<x_lammps_dummy_text>\s*units)")
    # ]


    # If force-field parameters appear in the files that will be read, these commands tell LAMMPS what kinds of force fields are being used:
    # pair_style, bond_style, angle_style, dihedral_style, improper_style.

    # initialization_force_field = [
    #     SM(r"\s*(?P<x_lammps_dummy_text>pair_style)"),
    #     SM(r"\s*(?P<x_lammps_dummy_text>bond_style)"),
    #     SM(r"\s*(?P<x_lammps_dummy_text>angle_style)"),
    #     SM(r"\s*(?P<x_lammps_dummy_text>dihedral_style)"),
    #     SM(r"\s*(?P<x_lammps_dummy_text>improper_style)")
    # ]


    # Atom definition
    # There are 3 ways to define atoms in LAMMPS. Read them in from a data or restart file via the read_data or read_restart commands.
    # These files can contain molecular topology information. Or create atoms on a lattice (with no molecular topology), using these commands:
    # lattice, region, create_box, create_atoms.
    # The entire set of atoms can be duplicated to make a larger simulation using the replicate command.

    # atom_definitation = SM(
    #     name='atom_definitation',
    #     startReStr="",
    #     # forwardMatch=True,
    #     weak=True,
    #     subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         SM(r"\s*(?P<x_lammps_dummy_text>create_atoms)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>create_box)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>lattice)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>read_data)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>read_dump)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>read_restart)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>region)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>replicate)")
    #     ]
    #
    # )



    # Settings
    # Once atoms and molecular topology are defined, a variety of settings can be specified: force field coefficients, simulation parameters, output options, etc.
    # Settings:

    # settings_other = SM(
    #     name='lammps-input-header',
    #     startReStr="",
    #     # forwardMatch=True,
    #     subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         SM(r"\s*(?P<x_lammps_dummy_text>comm_style)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>mass)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>set)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>velocity)")
    #     ]
    # )
    #

    # settings_force_field_coefficients = SM(
    #     name='lammps-input-header',
    #     startReStr="",
    #     # forwardMatch=True,
    #     subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         SM(r"\s*(?P<x_lammps_dummy_text>pair_coeff)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>bond_coeff)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>angle_coeff)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>dihedral_coeff)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>improper_coeff)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>kspace_style)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>kspace_modify)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>dielectric)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>special_bonds)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>pair_modify)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>pair_write)")
    #     ]
    # )

    # settings_simulation_parameters = SM(
    #     name='lammps-input-header',
    #     startReStr="",
    #     # forwardMatch=True,
    #     subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         SM(r"\s*(?P<x_lammps_dummy_text>neighbor)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>neigh_modify)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>group)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>timestep)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>reset_timestep)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>run_style)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>min_style)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>min_modify)"),
    #     ]
    # )

    # settings_fixes = SM(
    #     name='lammps-input-header',
    #     startReStr="",
    #     # forwardMatch=True,
    #     subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         SM(r"\s*(?P<x_lammps_dummy_text>fix)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>fix_modify)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>unfix)"),
    #     ]
    # )


    # settings_output_options = SM(
    #     name='lammps-input-header',
    #     startReStr="",
    #     # forwardMatch=True,
    #     subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         SM(r"\s*(?P<x_lammps_dummy_text>dump)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>dump image)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>dump_modify)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>dump movie)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>restart)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>thermo)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>thermo_modify)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>thermo_style)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>undump)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>write_data)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>write_dump)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>write_restart)"),
    #     ]
    # )

    # Run a simulation (actions)
    # A molecular dynamics simulation is run using the run command. Energy minimization (molecular statics) is performed using the minimize command. A parallel tempering (replica-exchange) simulation can be run using the temper command.

    # actions = SM(
    #     name='lammps-input-header',
    #     startReStr="",
    #     # forwardMatch=True,
    #     subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         SM(r"\s*(?P<x_lammps_dummy_text>delete_atoms)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>delete_bonds)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>displace_atoms)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>change_box)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>minimize)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>neb prd)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>rerun)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>run)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>temper)"),
    #     ]
    # )

    # computes = SM(
    #     name='lammps-input-header',
    #     startReStr="",
    #     # forwardMatch=True,
    #     subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         SM(r"\s*(?P<x_lammps_dummy_text>compute)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>compute_modify)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>uncompute)"),
    #     ]
    # )

    # miscellaneous = SM(
    #     name='lammps-input-header',
    #     startReStr="",
    #     # forwardMatch=True,
    #     subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         SM(r"\s*(?P<x_lammps_dummy_text>clear)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>echo)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>if)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>include)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>jump)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>label)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>log)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>next)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>print)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>shell)"),
    #         SM(r"\s*(?P<x_lammps_dummy_text>variable)"),
    #     ]
    # )

    #
    # header = SM(
    #     name='lammps-input-header',
    #     startReStr="",
    #     forwardMatch=True,
    #     sections=['section_topology'],
    #     subMatchers=[
    #         SM(r"\s*read_data\s(?P<x_lammps_data_file>{[file]})".format(nformat))
    #     ]
    # )

    ########################################
    # return main Parser


    initialization = SM(
        name="initialization",
        startReStr="",
        # endReStr = r"\s*Have a nice day\.",
        # repeats = True,
        # required=True,
        # forwardMatch=True,
        sections=['section_topology'],
        # forwardMatch=True,
        # weak=True,
        repeats=True,
        # subFlags=SM.SubFlags.Unordered,
        subMatchers=[
            SM(r"(?P<x_lammps_dummy_text>\s*dimension)"),
            SM(r"(?P<x_lammps_dummy_text>\s*atom_modify)"),
            SM(r"(?P<x_lammps_dummy_text>\s*atom_style)"),
            SM(r"(?P<x_lammps_dummy_text>\s*boundary)"),
            SM(r"(?P<x_lammps_dummy_text>\s*newton)"),
            SM(r"(?P<x_lammps_dummy_text>\s*processors)"),
            SM(r"(?P<x_lammps_dummy_text>\s*units)"),

            SM(r"\s*(?P<x_lammps_dummy_text>pair_style)"),
            SM(r"\s*(?P<x_lammps_dummy_text>bond_style)"),
            SM(r"\s*(?P<x_lammps_dummy_text>angle_style)"),
            SM(r"\s*(?P<x_lammps_dummy_text>dihedral_style)"),
            SM(r"\s*(?P<x_lammps_dummy_text>improper_style)"),

            SM(r"\s*(?P<x_lammps_dummy_text>create_atoms)"),
            SM(r"\s*(?P<x_lammps_dummy_text>create_box)"),
            SM(r"\s*(?P<x_lammps_dummy_text>lattice)"),
            SM(r"\s*(?P<x_lammps_dummy_text>read_data)"),
            SM(r"\s*(?P<x_lammps_dummy_text>read_dump)"),
            SM(r"\s*(?P<x_lammps_dummy_text>read_restart)"),
            SM(r"\s*(?P<x_lammps_dummy_text>region)"),
            SM(r"\s*(?P<x_lammps_dummy_text>replicate)"),

            SM(r"\s*(?P<x_lammps_dummy_text>comm_style)"),
            SM(r"\s*(?P<x_lammps_dummy_text>mass)"),
            SM(r"\s*(?P<x_lammps_dummy_text>set)"),
            SM(r"\s*(?P<x_lammps_dummy_text>velocity)"),

            SM(r"\s*(?P<x_lammps_dummy_text>pair_coeff)"),
            SM(r"\s*(?P<x_lammps_dummy_text>bond_coeff)"),
            SM(r"\s*(?P<x_lammps_dummy_text>angle_coeff)"),
            SM(r"\s*(?P<x_lammps_dummy_text>dihedral_coeff)"),
            SM(r"\s*(?P<x_lammps_dummy_text>improper_coeff)"),
            SM(r"\s*(?P<x_lammps_dummy_text>kspace_style)"),
            SM(r"\s*(?P<x_lammps_dummy_text>kspace_modify)"),
            SM(r"\s*(?P<x_lammps_dummy_text>dielectric)"),
            SM(r"\s*(?P<x_lammps_dummy_text>special_bonds)"),
            SM(r"\s*(?P<x_lammps_dummy_text>pair_modify)"),
            SM(r"\s*(?P<x_lammps_dummy_text>pair_write)"),

            SM(r"\s*(?P<x_lammps_dummy_text>neighbor)"),
            SM(r"\s*(?P<x_lammps_dummy_text>neigh_modify)"),
            SM(r"\s*(?P<x_lammps_dummy_text>group)"),
            SM(r"\s*(?P<x_lammps_dummy_text>timestep)"),
            SM(r"\s*(?P<x_lammps_dummy_text>reset_timestep)"),
            SM(r"\s*(?P<x_lammps_dummy_text>run_style)"),
            SM(r"\s*(?P<x_lammps_dummy_text>min_style)"),
            SM(r"\s*(?P<x_lammps_dummy_text>min_modify)"),

            SM(r"\s*(?P<x_lammps_dummy_text>fix)"),
            SM(r"\s*(?P<x_lammps_dummy_text>fix_modify)"),
            SM(r"\s*(?P<x_lammps_dummy_text>unfix)"),

            SM(r"\s*(?P<x_lammps_dummy_text>dump)"),
            SM(r"\s*(?P<x_lammps_dummy_text>dump image)"),
            SM(r"\s*(?P<x_lammps_dummy_text>dump_modify)"),
            SM(r"\s*(?P<x_lammps_dummy_text>dump movie)"),
            SM(r"\s*(?P<x_lammps_dummy_text>restart)"),
            SM(r"\s*(?P<x_lammps_dummy_text>thermo)"),
            SM(r"\s*(?P<x_lammps_dummy_text>thermo_modify)"),
            SM(r"\s*(?P<x_lammps_dummy_text>thermo_style)"),
            SM(r"\s*(?P<x_lammps_dummy_text>undump)"),
            SM(r"\s*(?P<x_lammps_dummy_text>write_data)"),
            SM(r"\s*(?P<x_lammps_dummy_text>write_dump)"),
            SM(r"\s*(?P<x_lammps_dummy_text>write_restart)"),
        ]
    )






    actions = SM(
        name='lammps-input-header',
        startReStr="",
        forwardMatch=True,
        sections=['section_topology'],
        subFlags=SM.SubFlags.Unordered,
        subMatchers=[
            SM(r"\s*(?P<x_lammps_dummy_text>delete_atoms)"),
            SM(r"\s*(?P<x_lammps_dummy_text>delete_bonds)"),
            SM(r"\s*(?P<x_lammps_dummy_text>displace_atoms)"),
            SM(r"\s*(?P<x_lammps_dummy_text>change_box)"),
            SM(r"\s*(?P<x_lammps_dummy_text>minimize)"),
            SM(r"\s*(?P<x_lammps_dummy_text>neb prd)"),
            SM(r"\s*(?P<x_lammps_dummy_text>rerun)"),
            SM(r"\s*(?P<x_lammps_dummy_text>run)"),
            SM(r"\s*(?P<x_lammps_dummy_text>temper)"),

            SM(r"\s*(?P<x_lammps_dummy_text>compute)"),
            SM(r"\s*(?P<x_lammps_dummy_text>compute_modify)"),
            SM(r"\s*(?P<x_lammps_dummy_text>uncompute)"),

            SM(r"\s*(?P<x_lammps_dummy_text>clear)"),
            SM(r"\s*(?P<x_lammps_dummy_text>echo)"),
            SM(r"\s*(?P<x_lammps_dummy_text>if)"),
            SM(r"\s*(?P<x_lammps_dummy_text>include)"),
            SM(r"\s*(?P<x_lammps_dummy_text>jump)"),
            SM(r"\s*(?P<x_lammps_dummy_text>label)"),
            SM(r"\s*(?P<x_lammps_dummy_text>log)"),
            SM(r"\s*(?P<x_lammps_dummy_text>next)"),
            SM(r"\s*(?P<x_lammps_dummy_text>print)"),
            SM(r"\s*(?P<x_lammps_dummy_text>shell)"),
            SM(r"\s*(?P<x_lammps_dummy_text>variable)"),
        ]
    )

    # test case: try to separate each basic steps in the input file
    # not yet working:
    # - if I try to close a section using endStr its eat my line
    # - if I try to catch the line with the next SM
    return SM(
        name="root",
        startReStr="",
        sections=['section_run'],
        fixedStartValues={'program_name': 'LAMMPS'},
        forwardMatch=True,
        weak=True,
        # subFlags=SM.SubFlags.Unordered,
        subMatchers=[
            SM(
                startReStr="",
                # endReStr=r"pair_style.*",
                # sections=['test'],
                forwardMatch=True,
                subFlags=SM.SubFlags.Unordered,
                # weak=True,
                repeats=True,
                subMatchers=[
                    SM(
                        r"(?P<x_lammps_dummy_text>a.*)",
                        repeats=True,
                    ),
                    SM(
                        r"(?P<x_lammps_dummy_text>b.*)",
                        repeats=True,
                    )
                ]
            ),
            # SM(r"(?P<x_lammps_dummy_text>pair_style.*)"),
            SM(
                startReStr="",
                # endReStr=r"pair_style.*",
                forwardMatch=True,
                subFlags=SM.SubFlags.Unordered,
                # weak=True,
                repeats=True,
                subMatchers=[

                    SM(r"(?P<x_lammps_dummy_text>pair_style.*)")
                ]
            )
        ]
    )

    # return SM(
    #     name='Root',
    #     startReStr="",
    #     forwardMatch=True,
    #     weak=True,
    #     sections=['section_run'],
    #     fixedStartValues={'program_name': 'LAMMPS'},
    #     # subFlags=SM.SubFlags.Unordered,
    #     subMatchers=[
    #         initialization,
    #         actions
    #         # SM(
    #         #     name="atom description",
    #         #     startReStr="",
    #         #     # forwardMatch=True,
    #         #     # weak=True,
    #         #     sections=['section_topology'],
    #         #     subMatchers=[
    #         #         atom_definitation
    #         #     ]
    #         # ),
    #         # SM(
    #         #     name="repetable section",
    #         #     startReStr="",
    #         #     forwardMatch=True,
    #         #     subMatchers=[
    #         #         settings_simulation_parameters,
    #         #         settings_fixes,
    #         #         settings_force_field_coefficients,
    #         #         settings_output_options,
    #         #         settings_other,
    #         #         actions
    #         #     ]
    #         # )
    #     ]
    # )
    #
    # # return SM(
    # #     name='root',
    # #     # repeats=True,
    # #     # required=True,
    # #     forwardMatch=True,
    # #     startReStr="",
    # #     fixedStartValues={'program_name': 'LAMMPS'},
    # #     sections=['section_run'],
    # #     subMatchers=[
    # #         header
    # #     ]
    # # )
    # # return SM(
    # #     name='root',
    # #     repeats=True,
    # #     required=True,
    # #     forwardMatch=True,
    # #     startReStr="",
    # #     subMatchers=[
    # #         SM(
    # #             name='newRun',
    # #             startReStr=r"",
    # #             repeats=True,
    #             # required=True,
    #             # forwardMatch=True,
    #             # sections=['section_run'],
    #             # subFlags=SM.SubFlags.Unordered,
    #             # fixedStartValues={'program_name': 'LAMMPS'},
    #             # subMatchers=[
    #             # ]),
    #         # SM(r"read_data\s(?P<x_lammps_dummy>{[file]})".format(nformat),
    #         #    sections=['section_topology'])
    #     # ])


def build_build_LammpsInputFileKeywordsSimpleMatcher():
    """Builds the list of SimpleMatchers to parse the control.in keywords of FHI-aims.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       List of SimpleMatchers that parses control.in keywords of FHI-aims.
    """
    # Now follows the list to match the keywords from the control.in.
    # Explicitly add ^ to ensure that the keyword is not within a comment.
    # Repating occurrences of the same keywords are captured.
    # List the matchers in alphabetical order according to keyword name.
    #
    return [
        # Initialization
        # Set parameters that need to be defined before atoms are created or read-in from a file.
        SM(r"(?P<x_lammps_dummy_text>\s*dimension)", repeats=True),
        SM(r"(?P<x_lammps_dummy_text>\s*atom_modify)", repeats=True),
        SM(r"(?P<x_lammps_dummy_text>\s*atom_style)", repeats=True),
        SM(r"(?P<x_lammps_dummy_text>\s*boundary)", repeats=True),
        SM(r"(?P<x_lammps_dummy_text>\s*newton)", repeats=True),
        SM(r"(?P<x_lammps_dummy_text>\s*processors)", repeats=True),

        # Units: lj, real, metal, si, cgs, electron, micro,nano
        SM(r"\s*units\s(?P<x_lammps_input_units_store>\w+)", repeats=True),

        # If force-field parameters appear in the files that will be read, these commands tell LAMMPS what kinds of force fields are being used:
        # pair_style, bond_style, angle_style, dihedral_style, improper_style.
        SM(r"\s*(?P<x_lammps_dummy_text>pair_style)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>bond_style)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>angle_style)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>dihedral_style)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>improper_style)", repeats=True),

        # Atom definition
        # There are 3 ways to define atoms in LAMMPS. Read them in from a data or restart file via the read_data or read_restart commands.
        # These files can contain molecular topology information. Or create atoms on a lattice (with no molecular topology), using these commands:
        # lattice, region, create_box, create_atoms.
        # The entire set of atoms can be duplicated to make a larger simulation using the replicate command.

        SM(r"\s*(?P<x_lammps_dummy_text>create_atoms)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>create_box)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>lattice)", repeats=True),
        SM(r"\s*read_data\s(?P<x_lammps_data_file_store>{[file]})".format(nformat), repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>read_dump)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>read_restart)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>region)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>replicate)", repeats=True),

        # Settings
        # Once atoms and molecular topology are defined, a variety of settings can be specified: force field coefficients, simulation parameters, output options, etc.
        # settings_other:

        SM(r"\s*(?P<x_lammps_dummy_text>comm_style)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>mass)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>set)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>velocity)", repeats=True),

        # settings_force_field_coefficients
        SM(r"\s*(?P<x_lammps_dummy_text>pair_coeff)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>bond_coeff)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>angle_coeff)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>dihedral_coeff)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>improper_coeff)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>kspace_style)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>kspace_modify)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>dielectric)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>special_bonds)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>pair_modify)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>pair_write)", repeats=True),

        # settings_simulation_parameters
        SM(r"\s*(?P<x_lammps_dummy_text>neighbor)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>neigh_modify)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>group)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>timestep)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>reset_timestep)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>run_style)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>min_style)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>min_modify)", repeats=True),

        # settings_fixes
        SM(r"\s*(?P<x_lammps_dummy_text>fix)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>fix_modify)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>unfix)", repeats=True),

        # settings_output_options
        SM(r"\s*(?P<x_lammps_dummy_text>dump)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>dump image)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>dump_modify)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>dump movie)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>restart)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>thermo)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>thermo_modify)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>thermo_style)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>undump)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>write_data)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>write_dump)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>write_restart)", repeats=True),

        # computes
        SM(r"\s*(?P<x_lammps_dummy_text>compute)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>compute_modify)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>uncompute)", repeats=True),


        # Run a simulation (actions)
        # A molecular dynamics simulation is run using the run command. Energy minimization (molecular statics) is performed using the minimize command. A parallel tempering (replica-exchange) simulation can be run using the temper command.
        SM(r"\s*(?P<x_lammps_dummy_text>delete_atoms)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>delete_bonds)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>displace_atoms)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>change_box)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>minimize)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>neb prd)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>rerun)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>run)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>temper)", repeats=True),

        # miscellaneous
        SM(r"\s*(?P<x_lammps_dummy_text>clear)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>echo)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>if)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>include)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>jump)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>label)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>log)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>next)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>print)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>shell)", repeats=True),
        SM(r"\s*(?P<x_lammps_dummy_text>variable)", repeats=True),

        #
        # SM (r"^\s*charge\s+(?P<fhi_aims_controlIn_charge>[-+0-9.eEdD]+)", repeats = True),
        # # need to distinguish different cases
        # SM (r"^\s*occupation_type\s+",
        #     forwardMatch = True,
        #     repeats = True,
        #     subMatchers = [
        #         SM (r"^\s*occupation_type\s+(?P<fhi_aims_controlIn_occupation_type>[-_a-zA-Z]+)\s+(?P<fhi_aims_controlIn_occupation_width>[-+0-9.eEdD]+)\s+(?P<fhi_aims_controlIn_occupation_order>[0-9]+)"),
        #         SM (r"^\s*occupation_type\s+(?P<fhi_aims_controlIn_occupation_type>[-_a-zA-Z]+)\s+(?P<fhi_aims_controlIn_occupation_width>[-+0-9.eEdD]+)")
        #     ]),
        # # using separate sections
        # SM (r"^\s*species\s*(?P<fhi_aims_controlIn_species_name>[a-zA-Z]+)",
        #     sections = ["fhi_aims_section_controlIn_basis_set"],
        #     repeats=True,
        #     subFlags = SM.SubFlags.Unordered,
        #     subMatchers = [
        #         SM (r"\s*\#\s*Definition of \"minimal\" basis",
        #             repeats=True,
        #             subFlags = SM.SubFlags.Unordered,
        #             subMatchers = [
        #                 SM (r"^\s*(?P<fhi_aims_controlIn_basis_func_type>[-_a-zA-Z0-9]+)"
        #                     "\s*(?P<fhi_aims_controlIn_basis_func_n>[0-9]+)"
        #                     "\s+(?P<fhi_aims_controlIn_basis_func_l>[a-zA-Z])"
        #                     "\s+(?P<fhi_aims_controlIn_basis_func_radius>[.0-9]+)",
        #                     repeats = True),
        #                 SM (r"^\s*(?P<fhi_aims_controlIn_basis_func_type>[-_a-zA-Z0-9]+)"
        #                     "\s*(?P<fhi_aims_controlIn_basis_func_n>[0-9]+)"
        #                     "\s+(?P<fhi_aims_controlIn_basis_func_l>[a-zA-Z])\s*auto",
        #                     repeats = True)
        #             ])
        #     ])
    ]

def build_LammpsInputFileSimpleMatcher2():
    """Builds the SimpleMatcher to parse the control.in file of FHI-aims.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses control.in file of FHI-aims.
    """
    return SM(
        name = 'Root1',
        startReStr = "",
        sections = ['section_run'],
        forwardMatch = True,
        weak = True,
        subMatchers = [
            SM(
                name = 'Root2',
                startReStr = "",
                sections = ['section_topology'],
                forwardMatch = True,
                weak = True,
                # The search is done unordered since the keywords do not appear in a specific order.
                subFlags = SM.SubFlags.Unordered,
                subMatchers = build_build_LammpsInputFileKeywordsSimpleMatcher()
            )
        ])



def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections dos, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
        'section_dos': CachingLvl,
        'section_run': CachingLvl,
        'section_single_configuration_calculation': CachingLvl,
    }
    # Set all dos metadata to Cache as they need post-processsing.
    for name in metaInfoEnv.infoKinds:
        if name.endswith('_store'):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the FHI-aims DOS file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections dos, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get dos file description
    LammpsInputSimpleMatcher = build_LammpsInputFileSimpleMatcher2()

    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../../../nomad-meta-info/meta_info/nomad_meta_info/lammps.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)

    # set parser info
    parserInfo = {'name': 'lammps-input-parser', 'version': '1.0'}

    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)

    # start parsing
    mainFunction(
        mainFileDescription=LammpsInputSimpleMatcher,
        metaInfoEnv=metaInfoEnv,
        parserInfo=parserInfo,
        cachingLevelForMetaName=cachingLevelForMetaName,
        superContext=LammpsInputParserContext()
    )


if __name__ == "__main__":
    main(CachingLevel.Forward)



# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from LAMMPS.BaseClasses import abstract_cmd, abstract_cmd_arg


def read_data(filename, *args):
    """Wrapper for read_data command"""


    class add(abstract_cmd_arg):
        """add arg = append or Nstart or merge
                append = add new atoms with IDs appended to current IDs
                Nstart = add new atoms with IDs starting with Nstart
                merge = add new atoms with their IDs unchanged"""

        def __init__(self, arg):
            self.arg = arg

    class offset(abstract_cmd_arg):
        """offset args = toff boff aoff doff ioff
                toff = offset to add to atom types
                boff = offset to add to bond types
                aoff = offset to add to angle types
                doff = offset to add to dihedral types
                ioff = offset to add to improper types"""

        def __init__(self, toff, boff, aoff, doff, ioff):
            self.toff = toff
            self.boff = boff
            self.aoff = aoff
            self.doff = doff
            self.ioff = ioff

    class shift(abstract_cmd_arg):
        """shift args = Sx Sy Sz
                Sx,Sy,Sz = distance to shift atoms when adding to system (distance units)"""

        def __init__(self, Sx, Sy, Sz):
            self.Sx = Sx
            self.Sy = Sy
            self.Sz = Sz

    class group(abstract_cmd_arg):
        """group args = groupID
                groupID = add atoms in data file to this group"""
        def __init__(self, group_id):
            self.group_id = group_id

    class nocoeff(abstract_cmd_arg):
        """nocoeff = ignore force field parameters"""

    class fix(abstract_cmd_arg):
        """fix args = fix-ID header-string section-string
                fix-ID = ID of fix to process header lines and sections of data file
                header-string = header lines containing this string will be passed to fix
                section-string = section names with this string will be passed to fix"""

        def __init__(self, id, header, section):
            self.id = id
            self.header = header
            self.section = section


    class extra_atom_types(abstract_cmd_arg):
        """extra/atom/types arg = # of extra atom types"""
        pass
    
    class extra_bond_types(abstract_cmd_arg):
        """extra/bond/types arg = # of extra bond types"""
        pass
    
    class extra_angle_types(abstract_cmd_arg):
        """extra/angle/types arg = # of extra angle types"""
        pass
    
    class extra_dihedral_types(abstract_cmd_arg):
        """extra/dihedral/types arg = # of extra dihedral types"""
        pass
    
    class extra_improper_types(abstract_cmd_arg):
        """extra/improper/types arg = # of extra improper types"""
        pass
    

if __name__ == '__main__':

    read_data('data.lj')
    read_data('../run7/data.polymer.gz')
    read_data('data.protein', 'fix', 'mycmap', 'crossterm' 'CMAP')
    read_data('data.water', 'add', 'append', 'offset', 3, 1, 1, 1, 1, 'shift', 0.0, 0.0, 50.0)
    # data = read_data('data.water')
    # data.add('append')
    # data.offset(3, 1, 1, 1, 1)
    # data.shift( 0.0, 0.0, 50.0)

    read_data('data.water', 'add', 'merge', 1, 'group', 'solvent')



# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module constructs the parser for the RIRPA module from TurboMole"""

import logging
import re
from nomadcore.simple_parser import SimpleMatcher as SM
from turbomoleparser.TurbomoleCommon import RE_FLOAT
import turbomoleparser.TurbomoleCommon as Common

logger = logging.getLogger("nomad.turbomoleParser")


class RIRPAparser(object):

    def __init__(self, context, key="rirpa"):
        context[key] = self
        self.__context = context
        self.__backend = None
        self.__no_hxx = False

    def purge_data(self):
        self.__no_hxx = False

    def set_backend(self, backend):
        self.__backend = backend

    def build_parser(self):
        references = SM(r"\s{5,}[^+ ]+",
                        name="references",
                        coverageIgnore=True,
                        repeats=True,
                        )
        header = SM(r"\s*\*+\s*PROGRAM\s+RIRPA\s*\*+\s*$",
                    name="header",
                    coverageIgnore=True,
                    subMatchers=[references],
                    endReStr=r"\s*\+-+\+"
                    )

        sub_matchers = [
            header,
            self.__context["geo"].build_qm_geometry_matcher(),
            self.__context["geo"].build_orbital_basis_matcher(
                "\s*\*\s*BASIS\s+SET\s+information:\s*$"),
            self.__context["geo"].build_auxiliary_basis_matcher(
                "\s*\*\s*AUXILIARY\s+BASIS\s+SET\s+information:\s*$"),
            # FIXME: figure out the difference between auxbasis and RI-J auxbasis
            self.__context["geo"].build_auxiliary_basis_matcher(
                "\s*\*\s*RIJ\s+AUXILIARY\s+BASIS\s+SET\s+information:\s*$"),
            self.__build_rpa_total_energy_matcher(),
            self.__context["gradient"].build_gradient_matcher(),
            Common.build_profiling_matcher(r"\s*Rirpa\s+profiling\s+cpu\s+wall\s+ratio\s*$"),
        ]

        return self.__context.build_module_matcher("rirpa", sub_matchers, "RIRPA")

    def __build_no_hxx_matcher(self):
        def set_flag(backend, groups):
            self.__no_hxx = True

        return SM(r"\s*The\s+HXX\s+energy\s+will\s+NOT\s+be\s+computed\s+due\s+to\s+nohxx"
                  r"\s+option.\s*$",
                  name="nohxx flag",
                  startReAction=set_flag
                  )

    def __build_rpa_total_energy_matcher(self):
        def get_total_energy(backend, groups):
            if not self.__no_hxx:
                backend.addRealValue("energy_total", float(groups[0]),
                                     self.__context.index_configuration(), unit="hartree")
            backend.addValue("electronic_structure_method", "RPA", self.__context.index_method())

        def get_correlation_energy(backend, groups):
            backend.addRealValue("energy_current", float(groups[0]),
                                 self.__context.index_configuration(), unit="hartree")

        def get_hf_energy(backend, groups):
            e_hf = float(groups[0])
            if abs(e_hf) < 1E-10:
                return
            index_config = backend.openSection("section_single_configuration_calculation")
            index_method = backend.openSection("section_method")
            references = {"section_method": index_method}
            backend.addValue("single_configuration_to_calculation_method_ref",
                             index_method, index_config)
            backend.addValue("single_configuration_calculation_to_system_ref",
                             self.__context.index_system(), index_config)
            backend.addValue("electronic_structure_method", "DFT", index_method)
            backend.addValue("calculation_method_kind", "absolute", index_method)
            index_xc = backend.openSection("section_XC_functionals")
            backend.setSectionInfo("section_XC_functionals", index_xc, references)
            backend.addValue('XC_functional_name', "HF_X", index_xc)
            backend.closeSection("section_XC_functionals", index_xc)
            backend.addRealValue("energy_total", e_hf, index_method, unit="hartree")

            index_link = backend.openSection("section_calculation_to_calculation_refs")
            references = {"section_single_configuration_calculation":
                              self.__context.index_configuration()}
            backend.setSectionInfo("section_calculation_to_calculation_refs", index_link,
                                   references)
            backend.addValue("calculation_to_calculation_kind", "source_calculation", index_link)
            backend.addValue("calculation_to_calculation_ref", index_config, index_link)
            backend.closeSection("section_calculation_to_calculation_refs", index_link)

            backend.closeSection("section_single_configuration_calculation", index_config)
            backend.closeSection("section_method", index_method)

        total_energy = SM(r"\s*\|\s*HXX\+RIRPA\s+total\s+energy\s*=\s*("+RE_FLOAT+")\s*\|\s*$",
                          name="RPA total energy",
                          startReAction=get_total_energy
                          )
        correlation = SM(r"\s*:\s*RIRPA\s+correlation\s+energy\s*=\s*("+RE_FLOAT+")\s*:\s*$",
                         name="RPA correlation ",
                         startReAction=get_correlation_energy
                         )
        hf_energy = SM(r"\s*:\s*HXX\s+total\s+energy\s*=\s*("+RE_FLOAT+")\s*:\s*$",
                       name="RPA total energy",
                       startReAction=get_hf_energy
                       )

        return SM(r"\s*Complex\s+frequency\s+integration\s*$",
                  subMatchers=[
                      total_energy,
                      correlation,
                      hf_energy
                  ]
                  )



from optimade.models import StructureResource
from optimade.adapters.base import EntryAdapter

from .aiida import get_aiida_structure_data
from .ase import get_ase_atoms
from .cif import get_cif
from .proteindatabank import get_pdb, get_pdbx_mmcif
from .pymatgen import get_pymatgen


__all__ = ("Structure",)


class Structure(EntryAdapter):
    """Lazy structure resource converter
    :param structure: a single JSON OPTIMADE structure resource entry.
    """

    ENTRY_RESOURCE = StructureResource
    _type_converters = {
        "aiida_structuredata": get_aiida_structure_data,
        "ase": get_ase_atoms,
        "cif": get_cif,
        "pdb": get_pdb,
        "pdbx_mmcif": get_pdbx_mmcif,
        "pymatgen": get_pymatgen,
    }
